@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2024 The OpenLDAP Foundation.
+ * Copyright 1998-2026 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 1998-2001 Net Boolean Incorporated.
  * Portions Copyright 2001-2003 IBM Corporation.
@@ -2061,10 +2061,12 @@ static void print_reference(
 	}
 
 	if( refs ) {
-		int i;
-		for( i=0; refs[i] != NULL; i++ ) {
-			tool_write_ldif( ldif ? LDIF_PUT_COMMENT : LDIF_PUT_VALUE,
-				"ref", refs[i], strlen(refs[i]) );
+		if( ldif < 2 ) {
+			int i;
+			for( i=0; refs[i] != NULL; i++ ) {
+				tool_write_ldif( ldif ? LDIF_PUT_COMMENT : LDIF_PUT_VALUE,
+					"ref", refs[i], strlen(refs[i]) );
+			}
 		}
 		ber_memvfree( (void **) refs );
 	}
