@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2024 The OpenLDAP Foundation.
+ * Copyright 1998-2026 The OpenLDAP Foundation.
  * Portions Copyright 2007 by Howard Chu, Symas Corporation.
  * All rights reserved.
  *
@@ -905,7 +905,8 @@ lload_listener(
 #endif /* SO_KEEPALIVE || TCP_NODELAY */
 
     if ( sl->sl_is_proxied ) {
-        if ( !proxyp( s, from ) ) {
+		ber_socklen_t dummy;
+        if ( !proxyp( s, from, &dummy ) ) {
             Debug( LDAP_DEBUG_ANY, "lload_listener: "
                     "proxyp(%ld) failed\n",
                     (long)s );
