@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2011-2021 The OpenLDAP Foundation.
+ * Copyright 2011-2026 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -58,6 +58,8 @@ static MDB_cursor *mcp = NULL, *mcd = NULL;
 static MDB_val key, data;
 static ID previd = NOID;
 
+static int reindexing;
+
 typedef struct dn_id {
 	ID id;
 	struct berval dn;
@@ -74,10 +76,11 @@ static Filter		*tool_filter;
 static Entry		*tool_next_entry;
 
 static ID mdb_tool_ix_id;
-static Operation *mdb_tool_ix_op;
+static BackendDB *mdb_tool_ix_be;
 static MDB_txn *mdb_tool_ix_txn;
 static int mdb_tool_index_tcount, mdb_tool_threads;
 static IndexRec *mdb_tool_index_rec;
+static AttrIxInfo **mdb_tool_axinfo;
 static struct mdb_info *mdb_tool_info;
 static ldap_pvt_thread_mutex_t mdb_tool_index_mutex;
 static ldap_pvt_thread_cond_t mdb_tool_index_cond_main;
@@ -100,6 +103,9 @@ mdb_tool_entry_get_int( BackendDB *be, ID id, Entry **ep );
 int mdb_tool_entry_open(
 	BackendDB *be, int mode )
 {
+	if ( slapMode & SLAP_TOOL_DRYRUN )
+		return 0;
+
 	/* In Quick mode, commit once per 500 entries */
 	mdb_writes = 0;
 	if ( slapMode & SLAP_TOOL_QUICK )
@@ -107,6 +113,7 @@ int mdb_tool_entry_open(
 	else
 		mdb_writes_per_commit = 1;
 
+#ifdef MDB_TOOL_IDL_CACHING			/* threaded indexing has no performance advantage */
 	/* Set up for threaded slapindex */
 	if (( slapMode & (SLAP_TOOL_QUICK|SLAP_TOOL_READONLY)) == SLAP_TOOL_QUICK ) {
 		if ( !mdb_tool_info ) {
@@ -116,12 +123,16 @@ int mdb_tool_entry_open(
 			ldap_pvt_thread_cond_init( &mdb_tool_index_cond_work );
 			if ( mdb->mi_nattrs ) {
 				int i;
-#if 0			/* threaded indexing has no performance advantage */
 				mdb_tool_threads = slap_tool_thread_max - 1;
-#endif
 				if ( mdb_tool_threads > 1 ) {
 					mdb_tool_index_rec = ch_calloc( mdb->mi_nattrs, sizeof( IndexRec ));
+					mdb_tool_axinfo = ch_calloc( mdb_tool_threads, sizeof( AttrIxInfo* ) +
+						sizeof( AttrIxInfo ));
+					mdb_tool_axinfo[0] = (AttrIxInfo *)(mdb_tool_axinfo + mdb_tool_threads);
+					for (i=1; i<mdb_tool_threads; i++)
+						mdb_tool_axinfo[i] = mdb_tool_axinfo[i-1]+1;
 					mdb_tool_index_tcount = mdb_tool_threads - 1;
+					mdb_tool_ix_be = be;
 					for (i=1; i<mdb_tool_threads; i++) {
 						int *ptr = ch_malloc( sizeof( int ));
 						*ptr = i;
@@ -133,6 +144,7 @@ int mdb_tool_entry_open(
 			}
 		}
 	}
+#endif
 
 	return 0;
 }
@@ -140,7 +152,12 @@ int mdb_tool_entry_open(
 int mdb_tool_entry_close(
 	BackendDB *be )
 {
+	if ( slapMode & SLAP_TOOL_DRYRUN )
+		return 0;
+
+#ifdef MDB_TOOL_IDL_CACHING
 	if ( mdb_tool_info ) {
+		int i;
 		slapd_shutdown = 1;
 		ldap_pvt_thread_mutex_lock( &mdb_tool_index_mutex );
 
@@ -164,7 +181,22 @@ int mdb_tool_entry_close(
 		slapd_shutdown = 0;
 		ch_free( mdb_tool_index_rec );
 		mdb_tool_index_tcount = mdb_tool_threads - 1;
+		if (mdb_tool_txn)
+			MDB_TOOL_IDL_FLUSH( be, mdb_tool_txn );
+		for (i=0; i<mdb_tool_threads; i++) {
+			mdb_tool_idl_cache *ic;
+			mdb_tool_idl_cache_entry *ice;
+			while ((ic = mdb_tool_axinfo[i]->ai_clist)) {
+				mdb_tool_axinfo[i]->ai_clist = ic->head;
+				free(ic);
+			}
+			while ((ice = mdb_tool_axinfo[i]->ai_flist)) {
+				mdb_tool_axinfo[i]->ai_flist = ice->next;
+				free(ice);
+			}
+		}
 	}
+#endif
 
 	if( idcursor ) {
 		mdb_cursor_close( idcursor );
@@ -184,7 +216,6 @@ int mdb_tool_entry_close(
 	}
 	if( mdb_tool_txn ) {
 		int rc;
-		MDB_TOOL_IDL_FLUSH( be, mdb_tool_txn );
 		if (( rc = mdb_txn_commit( mdb_tool_txn ))) {
 			Debug( LDAP_DEBUG_ANY,
 				LDAP_XSTRING(mdb_tool_entry_close) ": database %s: "
@@ -194,6 +225,20 @@ int mdb_tool_entry_close(
 		}
 		mdb_tool_txn = NULL;
 	}
+	if( reindexing ) {
+		struct mdb_info *mdb = be->be_private;
+		if ( !txi ) {
+			int rc = mdb_txn_begin( mdb->mi_dbenv, NULL, 0, &txi );
+			if( rc != 0 ) {
+				Debug( LDAP_DEBUG_ANY,
+					"=> " LDAP_XSTRING(mdb_tool_entry_close) ": database %s: "
+					"txn_begin failed: %s (%d)\n",
+					be->be_suffix[0].bv_val, mdb_strerror(rc), rc );
+				return -1;
+			}
+		}
+		mdb_drop( txi, mdb->mi_idxckp, 0 );
+	}
 	if( txi ) {
 		int rc;
 		if (( rc = mdb_txn_commit( txi ))) {
@@ -390,13 +435,15 @@ mdb_tool_entry_get_int( BackendDB *be, ID id, Entry **ep )
 			}
 		}
 	}
-	rc = mdb_entry_decode( &op, mdb_tool_txn, &data, &e );
+	rc = mdb_entry_decode( &op, mdb_tool_txn, &data, id, &e );
 	e->e_id = id;
 	if ( !BER_BVISNULL( &dn )) {
 		e->e_name = dn;
 		e->e_nname = ndn;
 	} else {
+		e->e_name.bv_len = 0;
 		e->e_name.bv_val = NULL;
+		e->e_nname.bv_len = 0;
 		e->e_nname.bv_val = NULL;
 	}
 
@@ -484,7 +531,7 @@ static int mdb_tool_next_id(
 				"next_id failed: %s (%d)",
 				mdb_strerror(rc), rc );
 		Debug( LDAP_DEBUG_ANY,
-			"=> mdb_tool_next_id: %s\n", text->bv_val, 0, 0 );
+			"=> mdb_tool_next_id: %s\n", text->bv_val );
 			return rc;
 		}
 		rc = mdb_dn2id_add( op, mcp, mcd, pid, 1, 1, e );
@@ -493,7 +540,7 @@ static int mdb_tool_next_id(
 				"dn2id_add failed: %s (%d)",
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
-				"=> mdb_tool_next_id: %s\n", text->bv_val, 0, 0 );
+				"=> mdb_tool_next_id: %s\n", text->bv_val );
 		} else if ( hole ) {
 			MDB_val key, data;
 			if ( nholes == nhmax - 1 ) {
@@ -519,7 +566,7 @@ static int mdb_tool_next_id(
 					"dummy id2entry add failed: %s (%d)",
 					mdb_strerror(rc), rc );
 				Debug( LDAP_DEBUG_ANY,
-					"=> mdb_tool_next_id: %s\n", text->bv_val, 0, 0 );
+					"=> mdb_tool_next_id: %s\n", text->bv_val );
 			}
 		}
 	} else if ( !hole ) {
@@ -577,7 +624,6 @@ mdb_tool_index_add(
 				return rc;
 		}
 		mdb_tool_ix_id = e->e_id;
-		mdb_tool_ix_op = op;
 		mdb_tool_ix_txn = txn;
 		ldap_pvt_thread_mutex_lock( &mdb_tool_index_mutex );
 		/* Wait for all threads to be ready */
@@ -589,33 +635,37 @@ mdb_tool_index_add(
 		for ( i=1; i<mdb_tool_threads; i++ )
 			mdb_tool_index_rec[i].ir_i = LDAP_BUSY;
 		mdb_tool_index_tcount = mdb_tool_threads - 1;
-		ldap_pvt_thread_mutex_unlock( &mdb_tool_index_mutex );
 		ldap_pvt_thread_cond_broadcast( &mdb_tool_index_cond_work );
-
-		rc = mdb_index_recrun( op, txn, mdb, ir, e->e_id, 0 );
-		if ( rc )
-			return rc;
-		ldap_pvt_thread_mutex_lock( &mdb_tool_index_mutex );
-		for ( i=1; i<mdb_tool_threads; i++ ) {
-			if ( mdb_tool_index_rec[i].ir_i == LDAP_BUSY ) {
-				ldap_pvt_thread_cond_wait( &mdb_tool_index_cond_main,
-					&mdb_tool_index_mutex );
-				i--;
-				continue;
-			}
-			if ( mdb_tool_index_rec[i].ir_i ) {
-				rc = mdb_tool_index_rec[i].ir_i;
-				break;
-			}
-		}
 		ldap_pvt_thread_mutex_unlock( &mdb_tool_index_mutex );
-		return rc;
+
+		return mdb_index_recrun( op, txn, mdb, ir, e->e_id, 0 );
 	} else
 	{
 		return mdb_index_entry_add( op, txn, e );
 	}
 }
 
+static int
+mdb_tool_index_finish()
+{
+	int i, rc = 0;
+	ldap_pvt_thread_mutex_lock( &mdb_tool_index_mutex );
+	for ( i=1; i<mdb_tool_threads; i++ ) {
+		if ( mdb_tool_index_rec[i].ir_i == LDAP_BUSY ) {
+			ldap_pvt_thread_cond_wait( &mdb_tool_index_cond_main,
+				&mdb_tool_index_mutex );
+			i--;
+			continue;
+		}
+		if ( mdb_tool_index_rec[i].ir_i ) {
+			rc = mdb_tool_index_rec[i].ir_i;
+			break;
+		}
+	}
+	ldap_pvt_thread_mutex_unlock( &mdb_tool_index_mutex );
+	return rc;
+}
+
 ID mdb_tool_entry_put(
 	BackendDB *be,
 	Entry *e,
@@ -626,6 +676,9 @@ ID mdb_tool_entry_put(
 	Operation op = {0};
 	Opheader ohdr = {0};
 
+	if ( slapMode & SLAP_TOOL_DRYRUN )
+		return 0;
+
 	assert( be != NULL );
 	assert( slapMode & SLAP_TOOL_MODE );
 
@@ -634,7 +687,7 @@ ID mdb_tool_entry_put(
 	assert( text->bv_val[0] == '\0' );	/* overconservative? */
 
 	Debug( LDAP_DEBUG_TRACE, "=> " LDAP_XSTRING(mdb_tool_entry_put)
-		"( %ld, \"%s\" )\n", (long) e->e_id, e->e_dn, 0 );
+		"( %ld, \"%s\" )\n", (long) e->e_id, e->e_dn );
 
 	mdb = (struct mdb_info *) be->be_private;
 
@@ -646,7 +699,7 @@ ID mdb_tool_entry_put(
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-				 text->bv_val, 0, 0 );
+				 text->bv_val );
 			return NOID;
 		}
 	}
@@ -658,7 +711,7 @@ ID mdb_tool_entry_put(
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-				 text->bv_val, 0, 0 );
+				 text->bv_val );
 			return NOID;
 		}
 		if ( !mdb->mi_nextid ) {
@@ -672,7 +725,7 @@ ID mdb_tool_entry_put(
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-				 text->bv_val, 0, 0 );
+				 text->bv_val );
 			return NOID;
 		}
 		rc = mdb_cursor_open( mdb_tool_txn, mdb->mi_dn2id, &mcd );
@@ -682,7 +735,7 @@ ID mdb_tool_entry_put(
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-				 text->bv_val, 0, 0 );
+				 text->bv_val );
 			return NOID;
 		}
 	}
@@ -698,13 +751,16 @@ ID mdb_tool_entry_put(
 		goto done;
 	}
 
+	if ( mdb_tool_threads > 1 ) {
+		LDAP_SLIST_INSERT_HEAD( &op.o_extra, &mdb_tool_axinfo[0]->ai_oe, oe_next );
+	}
 	rc = mdb_tool_index_add( &op, mdb_tool_txn, e );
 	if( rc != 0 ) {
 		snprintf( text->bv_val, text->bv_len,
 				"index_entry_add failed: err=%d", rc );
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-			text->bv_val, 0, 0 );
+			text->bv_val );
 		goto done;
 	}
 
@@ -716,10 +772,13 @@ ID mdb_tool_entry_put(
 				"id2entry_add failed: err=%d", rc );
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-			text->bv_val, 0, 0 );
+			text->bv_val );
 		goto done;
 	}
 
+	if( mdb->mi_nattrs && mdb_tool_threads > 1 )
+		rc = mdb_tool_index_finish();
+
 done:
 	if( rc == 0 ) {
 		mdb_writes++;
@@ -739,7 +798,7 @@ done:
 						mdb_strerror(rc), rc );
 				Debug( LDAP_DEBUG_ANY,
 					"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-					text->bv_val, 0, 0 );
+					text->bv_val );
 				e->e_id = NOID;
 			}
 		}
@@ -758,7 +817,7 @@ done:
 			mdb_strerror(rc), rc );
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_put) ": %s\n",
-			text->bv_val, 0, 0 );
+			text->bv_val );
 		e->e_id = NOID;
 	}
 
@@ -780,7 +839,7 @@ int mdb_tool_entry_reindex(
 
 	Debug( LDAP_DEBUG_ARGS,
 		"=> " LDAP_XSTRING(mdb_tool_entry_reindex) "( %ld )\n",
-		(long) id, 0, 0 );
+		(long) id );
 	assert( tool_base == NULL );
 	assert( tool_filter == NULL );
 
@@ -798,6 +857,8 @@ int mdb_tool_entry_reindex(
 		return 0;
 	}
 
+	reindexing = 1;
+
 	/* Check for explicit list of attrs to index */
 	if ( adv ) {
 		int i, j, n;
@@ -831,7 +892,7 @@ int mdb_tool_entry_reindex(
 					Debug( LDAP_DEBUG_ANY,
 						LDAP_XSTRING(mdb_tool_entry_reindex)
 						": no index configured for %s\n",
-						adv[i]->ad_cname.bv_val, 0, 0 );
+						adv[i]->ad_cname.bv_val );
 					return -1;
 				}
 			}
@@ -845,7 +906,7 @@ int mdb_tool_entry_reindex(
 		Debug( LDAP_DEBUG_ANY,
 			LDAP_XSTRING(mdb_tool_entry_reindex)
 			": could not locate id=%ld\n",
-			(long) id, 0, 0 );
+			(long) id );
 		return -1;
 	}
 
@@ -855,7 +916,7 @@ int mdb_tool_entry_reindex(
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_reindex) ": "
 				"txn_begin failed: %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto done;
 		}
 	}
@@ -883,7 +944,7 @@ int mdb_tool_entry_reindex(
 
 	Debug( LDAP_DEBUG_TRACE,
 		"=> " LDAP_XSTRING(mdb_tool_entry_reindex) "( %ld )\n",
-		(long) id, 0, 0 );
+		(long) id );
 
 	op.o_hdr = &ohdr;
 	op.o_bd = be;
@@ -907,7 +968,7 @@ done:
 				Debug( LDAP_DEBUG_ANY,
 					"=> " LDAP_XSTRING(mdb_tool_entry_reindex)
 					": txn_commit failed: %s (%d)\n",
-					mdb_strerror(rc), rc, 0 );
+					mdb_strerror(rc), rc );
 				e->e_id = NOID;
 			}
 			mdb_cursor_close( cursor );
@@ -933,7 +994,7 @@ done:
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_reindex)
 			": txn_aborted! err=%d\n",
-			rc, 0, 0 );
+			rc );
 		e->e_id = NOID;
 		txi = NULL;
 	}
@@ -963,7 +1024,7 @@ ID mdb_tool_entry_modify(
 
 	Debug( LDAP_DEBUG_TRACE,
 		"=> " LDAP_XSTRING(mdb_tool_entry_modify) "( %ld, \"%s\" )\n",
-		(long) e->e_id, e->e_dn, 0 );
+		(long) e->e_id, e->e_dn );
 
 	mdb = (struct mdb_info *) be->be_private;
 
@@ -979,7 +1040,7 @@ ID mdb_tool_entry_modify(
 				mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_modify) ": %s\n",
-				 text->bv_val, 0, 0 );
+				 text->bv_val );
 			return NOID;
 		}
 	}
@@ -990,13 +1051,13 @@ ID mdb_tool_entry_modify(
 	op.o_tmpmfuncs = &ch_mfuncs;
 
 	/* id2entry index */
-	rc = mdb_id2entry_update( &op, mdb_tool_txn, NULL, e );
+	rc = mdb_id2entry_update( &op, mdb_tool_txn, idcursor, e );
 	if( rc != 0 ) {
 		snprintf( text->bv_val, text->bv_len,
 				"id2entry_update failed: err=%d", rc );
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_modify) ": %s\n",
-			text->bv_val, 0, 0 );
+			text->bv_val );
 		goto done;
 	}
 
@@ -1010,7 +1071,7 @@ done:
 					mdb_strerror(rc), rc );
 			Debug( LDAP_DEBUG_ANY,
 				"=> " LDAP_XSTRING(mdb_tool_entry_modify) ": "
-				"%s\n", text->bv_val, 0, 0 );
+				"%s\n", text->bv_val );
 			e->e_id = NOID;
 		}
 
@@ -1021,7 +1082,7 @@ done:
 			mdb_strerror(rc), rc );
 		Debug( LDAP_DEBUG_ANY,
 			"=> " LDAP_XSTRING(mdb_tool_entry_modify) ": %s\n",
-			text->bv_val, 0, 0 );
+			text->bv_val );
 		e->e_id = NOID;
 	}
 	mdb_tool_txn = NULL;
@@ -1030,12 +1091,181 @@ done:
 	return e->e_id;
 }
 
+int mdb_tool_entry_delete(
+	BackendDB *be,
+	struct berval *ndn,
+	struct berval *text )
+{
+	int rc;
+	struct mdb_info *mdb;
+	Operation op = {0};
+	Opheader ohdr = {0};
+	Entry *e;
+
+	assert( be != NULL );
+	assert( slapMode & SLAP_TOOL_MODE );
+
+	assert( text != NULL );
+	assert( text->bv_val != NULL );
+	assert( text->bv_val[0] == '\0' );	/* overconservative? */
+
+	assert ( ndn != NULL );
+	assert ( ndn->bv_val != NULL );
+
+	Debug( LDAP_DEBUG_TRACE,
+		"=> " LDAP_XSTRING(mdb_tool_entry_delete) "( %s )\n",
+		ndn->bv_val );
+
+	mdb = (struct mdb_info *) be->be_private;
+
+	assert( cursor == NULL );
+	if( cursor ) {
+		mdb_cursor_close( cursor );
+		cursor = NULL;
+	}
+	if( !mdb_tool_txn ) {
+		rc = mdb_txn_begin( mdb->mi_dbenv, NULL, 0, &mdb_tool_txn );
+		if( rc != 0 ) {
+			snprintf( text->bv_val, text->bv_len,
+				"txn_begin failed: %s (%d)",
+				mdb_strerror(rc), rc );
+			Debug( LDAP_DEBUG_ANY,
+				"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+				 text->bv_val );
+			return LDAP_OTHER;
+		}
+	}
+
+	rc = mdb_cursor_open( mdb_tool_txn, mdb->mi_dn2id, &cursor );
+	if( rc != 0 ) {
+		snprintf( text->bv_val, text->bv_len,
+			"cursor_open failed: %s (%d)",
+			mdb_strerror(rc), rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			 text->bv_val );
+		return LDAP_OTHER;
+	}
+
+	op.o_hdr = &ohdr;
+	op.o_bd = be;
+	op.o_tmpmemctx = NULL;
+	op.o_tmpmfuncs = &ch_mfuncs;
+
+	rc = mdb_dn2entry( &op, mdb_tool_txn, cursor, ndn, &e, NULL, 0 );
+	if( rc != 0 ) {
+		snprintf( text->bv_val, text->bv_len,
+			"dn2entry failed: %s (%d)",
+			mdb_strerror(rc), rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			 text->bv_val );
+		goto done;
+	}
+
+	/* check that we wouldn't orphan any children */
+	rc = mdb_dn2id_children( &op, mdb_tool_txn, e );
+	if( rc != MDB_NOTFOUND ) {
+		switch( rc ) {
+		case 0:
+			snprintf( text->bv_val, text->bv_len,
+				"delete failed:"
+				" subordinate objects must be deleted first");
+			break;
+		default:
+			snprintf( text->bv_val, text->bv_len,
+				"has_children failed: %s (%d)",
+				mdb_strerror(rc), rc );
+			break;
+		}
+		rc = -1;
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			 text->bv_val );
+		goto done;
+	}
+
+	/* delete from dn2id */
+	rc = mdb_dn2id_delete( &op, cursor, e->e_id, 1 );
+	if( rc != 0 ) {
+		snprintf( text->bv_val, text->bv_len,
+				"dn2id_delete failed: err=%d", rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			text->bv_val );
+		goto done;
+	}
+
+	/* deindex values */
+	rc = mdb_index_entry_del( &op, mdb_tool_txn, e );
+	if( rc != 0 ) {
+		snprintf( text->bv_val, text->bv_len,
+				"entry_delete failed: err=%d", rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			text->bv_val );
+		goto done;
+	}
+
+	/* do the deletion */
+	rc = mdb_id2entry_delete( be, mdb_tool_txn, e );
+	if( rc != 0 ) {
+		snprintf( text->bv_val, text->bv_len,
+				"id2entry_update failed: err=%d", rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			text->bv_val );
+		goto done;
+	}
+
+done:
+	/* free entry */
+	if( e != NULL ) {
+		mdb_entry_return( &op, e );
+	}
+
+	if( rc == 0 ) {
+		rc = mdb_txn_commit( mdb_tool_txn );
+		if( rc != 0 ) {
+			snprintf( text->bv_val, text->bv_len,
+					"txn_commit failed: %s (%d)",
+					mdb_strerror(rc), rc );
+			Debug( LDAP_DEBUG_ANY,
+				"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": "
+				"%s\n", text->bv_val );
+		}
+
+	} else {
+		mdb_txn_abort( mdb_tool_txn );
+		snprintf( text->bv_val, text->bv_len,
+			"txn_aborted! %s (%d)",
+			mdb_strerror(rc), rc );
+		Debug( LDAP_DEBUG_ANY,
+			"=> " LDAP_XSTRING(mdb_tool_entry_delete) ": %s\n",
+			text->bv_val );
+	}
+	mdb_tool_txn = NULL;
+	cursor = NULL;
+
+	return rc;
+}
+
 static void *
 mdb_tool_index_task( void *ctx, void *ptr )
 {
 	int base = *(int *)ptr;
+	Operation op = {0};
+	Opheader ohdr = {0};
+	AttrIxInfo ai = {0}, *aio;
 
 	free( ptr );
+	op.o_hdr = &ohdr;
+	op.o_bd = mdb_tool_ix_be;
+	op.o_tmpmemctx = NULL;
+	op.o_tmpmfuncs = &ch_mfuncs;
+	aio = mdb_tool_axinfo[base];
+	mdb_tool_axinfo[base] = &ai;
+	LDAP_SLIST_INSERT_HEAD( &op.o_extra, &ai.ai_oe, oe_next );
 	while ( 1 ) {
 		ldap_pvt_thread_mutex_lock( &mdb_tool_index_mutex );
 		mdb_tool_index_tcount--;
@@ -1047,11 +1277,13 @@ mdb_tool_index_task( void *ctx, void *ptr )
 			mdb_tool_index_tcount--;
 			if ( !mdb_tool_index_tcount )
 				ldap_pvt_thread_cond_signal( &mdb_tool_index_cond_main );
+			*aio = ai;
+			mdb_tool_axinfo[base] = aio;
 			ldap_pvt_thread_mutex_unlock( &mdb_tool_index_mutex );
 			break;
 		}
 		ldap_pvt_thread_mutex_unlock( &mdb_tool_index_mutex );
-		mdb_tool_index_rec[base].ir_i = mdb_index_recrun( mdb_tool_ix_op,
+		mdb_tool_index_rec[base].ir_i = mdb_index_recrun( &op,
 			mdb_tool_ix_txn,
 			mdb_tool_info, mdb_tool_index_rec, mdb_tool_ix_id, base );
 	}
@@ -1071,7 +1303,7 @@ mdb_tool_idl_cmp( const void *v1, const void *v2 )
 }
 
 static int
-mdb_tool_idl_flush_one( MDB_cursor *mc, AttrInfo *ai, mdb_tool_idl_cache *ic )
+mdb_tool_idl_flush_one( MDB_cursor *mc, AttrIxInfo *ai, mdb_tool_idl_cache *ic )
 {
 	mdb_tool_idl_cache_entry *ice;
 	MDB_val key, data[2];
@@ -1079,14 +1311,14 @@ mdb_tool_idl_flush_one( MDB_cursor *mc, AttrInfo *ai, mdb_tool_idl_cache *ic )
 	ID id, nid;
 
 	/* Freshly allocated, ignore it */
-	if ( !ic->head && ic->count <= MDB_IDL_DB_SIZE ) {
+	if ( !ic->head && ic->count <= MDB_idl_db_size ) {
 		return 0;
 	}
 
 	key.mv_data = ic->kstr.bv_val;
 	key.mv_size = ic->kstr.bv_len;
 
-	if ( ic->count > MDB_IDL_DB_SIZE ) {
+	if ( ic->count > MDB_idl_db_size ) {
 		while ( ic->flags & WAS_FOUND ) {
 			rc = mdb_cursor_get( mc, &key, data, MDB_SET );
 			if ( rc ) {
@@ -1135,25 +1367,19 @@ mdb_tool_idl_flush_one( MDB_cursor *mc, AttrInfo *ai, mdb_tool_idl_cache *ic )
 
 		data[0].mv_size = sizeof(ID);
 		rc = 0;
-		i = ic->offset;
 		for ( ice = ic->head, n=0; ice; ice = ice->next, n++ ) {
 			int end;
 			if ( ice->next ) {
 				end = IDBLOCK;
 			} else {
-				end = ic->count & (IDBLOCK-1);
+				end = (ic->count-ic->offset) & (IDBLOCK-1);
 				if ( !end )
 					end = IDBLOCK;
 			}
-			data[1].mv_size = end - i;
-			data[0].mv_data = &ice->ids[i];
-			i = 0;
-			rc = mdb_cursor_put( mc, &key, data, MDB_NODUPDATA|MDB_APPEND|MDB_MULTIPLE );
+			data[1].mv_size = end;
+			data[0].mv_data = ice->ids;
+			rc = mdb_cursor_put( mc, &key, data, MDB_APPENDDUP|MDB_MULTIPLE );
 			if ( rc ) {
-				if ( rc == MDB_KEYEXIST ) {
-					rc = 0;
-					continue;
-				}
 				rc = -1;
 				break;
 			}
@@ -1169,19 +1395,19 @@ mdb_tool_idl_flush_one( MDB_cursor *mc, AttrInfo *ai, mdb_tool_idl_cache *ic )
 }
 
 static int
-mdb_tool_idl_flush_db( MDB_txn *txn, AttrInfo *ai )
+mdb_tool_idl_flush_db( MDB_txn *txn, AttrInfo *ai, AttrIxInfo *ax )
 {
 	MDB_cursor *mc;
 	Avlnode *root;
 	int rc;
 
 	mdb_cursor_open( txn, ai->ai_dbi, &mc );
-	root = tavl_end( ai->ai_root, TAVL_DIR_LEFT );
+	root = ldap_tavl_end( ai->ai_root, TAVL_DIR_LEFT );
 	do {
-		rc = mdb_tool_idl_flush_one( mc, ai, root->avl_data );
+		rc = mdb_tool_idl_flush_one( mc, ax, root->avl_data );
 		if ( rc != -1 )
 			rc = 0;
-	} while ((root = tavl_next(root, TAVL_DIR_RIGHT)));
+	} while ((root = ldap_tavl_next(root, TAVL_DIR_RIGHT)));
 	mdb_cursor_close( mc );
 
 	return rc;
@@ -1196,8 +1422,8 @@ mdb_tool_idl_flush( BackendDB *be, MDB_txn *txn )
 
 	for ( i=0; i < mdb->mi_nattrs; i++ ) {
 		if ( !mdb->mi_attrs[i]->ai_root ) continue;
-		rc = mdb_tool_idl_flush_db( txn, mdb->mi_attrs[i] );
-		tavl_free(mdb->mi_attrs[i]->ai_root, NULL);
+		rc = mdb_tool_idl_flush_db( txn, mdb->mi_attrs[i], mdb_tool_axinfo[i % mdb_tool_threads] );
+		ldap_tavl_free(mdb->mi_attrs[i]->ai_root, NULL);
 		mdb->mi_attrs[i]->ai_root = NULL;
 		if ( rc )
 			break;
@@ -1215,13 +1441,14 @@ int mdb_tool_idl_add(
 	mdb_tool_idl_cache *ic, itmp;
 	mdb_tool_idl_cache_entry *ice;
 	int i, rc, lcount;
-	AttrInfo *ai = (AttrInfo *)mc;
+	AttrIxInfo *ax = (AttrIxInfo *)mc;
+	AttrInfo *ai = (AttrInfo *)ax->ai_ai;
 	mc = ai->ai_cursor;
 
 	dbi = ai->ai_dbi;
 	for (i=0; keys[i].bv_val; i++) {
 	itmp.kstr = keys[i];
-	ic = tavl_find( (Avlnode *)ai->ai_root, &itmp, mdb_tool_idl_cmp );
+	ic = ldap_tavl_find( ai->ai_root, &itmp, mdb_tool_idl_cmp );
 
 	/* No entry yet, create one */
 	if ( !ic ) {
@@ -1229,9 +1456,9 @@ int mdb_tool_idl_add(
 		ID nid;
 		int rc;
 
-		if ( ai->ai_clist ) {
-			ic = ai->ai_clist;
-			ai->ai_clist = ic->head;
+		if ( ax->ai_clist ) {
+			ic = ax->ai_clist;
+			ax->ai_clist = ic->head;
 		} else {
 			ic = ch_malloc( sizeof( mdb_tool_idl_cache ) + itmp.kstr.bv_len + 4 );
 		}
@@ -1243,8 +1470,8 @@ int mdb_tool_idl_add(
 		ic->count = 0;
 		ic->offset = 0;
 		ic->flags = 0;
-		tavl_insert( (Avlnode **)&ai->ai_root, ic, mdb_tool_idl_cmp,
-			avl_dup_error );
+		ldap_tavl_insert( &ai->ai_root, ic, mdb_tool_idl_cmp,
+			ldap_avl_dup_error );
 
 		/* load existing key count here */
 		key.mv_size = keys[i].bv_len;
@@ -1254,7 +1481,7 @@ int mdb_tool_idl_add(
 			ic->flags |= WAS_FOUND;
 			nid = *(ID *)data.mv_data;
 			if ( nid == 0 ) {
-				ic->count = MDB_IDL_DB_SIZE+1;
+				ic->count = MDB_idl_db_size+1;
 				ic->flags |= WAS_RANGE;
 			} else {
 				size_t count;
@@ -1267,14 +1494,14 @@ int mdb_tool_idl_add(
 		}
 	}
 	/* are we a range already? */
-	if ( ic->count > MDB_IDL_DB_SIZE ) {
+	if ( ic->count > MDB_idl_db_size ) {
 		ic->last = id;
 		continue;
 	/* Are we at the limit, and converting to a range? */
-	} else if ( ic->count == MDB_IDL_DB_SIZE ) {
+	} else if ( ic->count == MDB_idl_db_size ) {
 		if ( ic->head ) {
-			ic->tail->next = ai->ai_flist;
-			ai->ai_flist = ic->head;
+			ic->tail->next = ax->ai_flist;
+			ax->ai_flist = ic->head;
 		}
 		ic->head = ic->tail = NULL;
 		ic->last = id;
@@ -1282,11 +1509,11 @@ int mdb_tool_idl_add(
 		continue;
 	}
 	/* No free block, create that too */
-	lcount = ic->count & (IDBLOCK-1);
+	lcount = (ic->count-ic->offset) & (IDBLOCK-1);
 	if ( !ic->tail || lcount == 0) {
-		if ( ai->ai_flist ) {
-			ice = ai->ai_flist;
-			ai->ai_flist = ice->next;
+		if ( ax->ai_flist ) {
+			ice = ax->ai_flist;
+			ax->ai_flist = ice->next;
 		} else {
 			ice = ch_malloc( sizeof( mdb_tool_idl_cache_entry ));
 		}
@@ -1303,9 +1530,10 @@ int mdb_tool_idl_add(
 			ic->first = id;
 	}
 	ice = ic->tail;
-	if (!lcount || ice->ids[lcount-1] != id)
+	if (!lcount || ice->ids[lcount-1] != id) {
 		ice->ids[lcount] = id;
-	ic->count++;
+		ic->count++;
+	}
 	}
 
 	return 0;
@@ -1339,7 +1567,7 @@ mdb_dn2id_upgrade( BackendDB *be ) {
 
 	if (!(mi->mi_flags & MDB_NEED_UPGRADE)) {
 		Debug( LDAP_DEBUG_ANY, "database %s: No upgrade needed.\n",
-			be->be_suffix[0].bv_val, 0, 0 );
+			be->be_suffix[0].bv_val );
 		return 0;
 	}
 
@@ -1364,13 +1592,13 @@ mdb_dn2id_upgrade( BackendDB *be ) {
 	rc = mdb_txn_begin(mi->mi_dbenv, NULL, 0, &mt);
 	if (rc) {
 		Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_txn_begin failed, %s (%d)\n",
-			mdb_strerror(rc), rc, 0 );
+			mdb_strerror(rc), rc );
 		goto leave;
 	}
 	rc = mdb_cursor_open(mt, mi->mi_dbis[MDB_DN2ID], &mc);
 	if (rc) {
 		Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_open failed, %s (%d)\n",
-			mdb_strerror(rc), rc, 0 );
+			mdb_strerror(rc), rc );
 		goto leave;
 	}
 
@@ -1386,7 +1614,7 @@ mdb_dn2id_upgrade( BackendDB *be ) {
 		rc = mdb_cursor_get(mc, &key, &data, MDB_SET);
 		if (rc) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_get failed, %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto leave;
 		}
 		num[depth] = 1;
@@ -1394,7 +1622,7 @@ mdb_dn2id_upgrade( BackendDB *be ) {
 		rc = mdb_cursor_count(mc, &dkids);
 		if (rc) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_count failed, %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto leave;
 		}
 		if (dkids > 1) {
@@ -1422,7 +1650,7 @@ pop:
 		rc = mdb_cursor_get(mc, &key, &data, MDB_GET_BOTH);
 		if (rc) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_get(BOTH) failed, %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto leave;
 		}
 		data.mv_data = stack[depth].rdn;
@@ -1432,13 +1660,13 @@ pop:
 		rc = mdb_cursor_del(mc, 0);
 		if (rc) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_del failed, %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto leave;
 		}
 		rc = mdb_cursor_put(mc, &key, &data, 0);
 		if (rc) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_put failed, %s (%d)\n",
-				mdb_strerror(rc), rc, 0 );
+				mdb_strerror(rc), rc );
 			goto leave;
 		}
 		count++;
@@ -1459,25 +1687,25 @@ pop:
 			rc = mdb_txn_commit(mt);
 			if (rc) {
 				Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_txn_commit failed, %s (%d)\n",
-					mdb_strerror(rc), rc, 0 );
+					mdb_strerror(rc), rc );
 				goto leave;
 			}
 			rc = mdb_txn_begin(mi->mi_dbenv, NULL, 0, &mt);
 			if (rc) {
 				Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_txn_begin(2) failed, %s (%d)\n",
-					mdb_strerror(rc), rc, 0 );
+					mdb_strerror(rc), rc );
 				goto leave;
 			}
 			rc = mdb_cursor_open(mt, mi->mi_dbis[MDB_DN2ID], &mc);
 			if (rc) {
 				Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_open(2) failed, %s (%d)\n",
-					mdb_strerror(rc), rc, 0 );
+					mdb_strerror(rc), rc );
 				goto leave;
 			}
 			rc = mdb_cursor_get(mc, &key, &data, MDB_GET_BOTH);
 			if (rc) {
 				Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_cursor_get(2) failed, %s (%d)\n",
-					mdb_strerror(rc), rc, 0 );
+					mdb_strerror(rc), rc );
 				goto leave;
 			}
 			writes = 0;
@@ -1500,7 +1728,7 @@ leave:
 		r2 = mdb_txn_commit(mt);
 		if (r2) {
 			Debug(LDAP_DEBUG_ANY, "mdb_dn2id_upgrade: mdb_txn_commit(2) failed, %s (%d)\n",
-				mdb_strerror(r2), r2, 0 );
+				mdb_strerror(r2), r2 );
 			if (!rc)
 				rc = r2;
 		}
