@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2021 The OpenLDAP Foundation.
+ * Copyright 2001-2026 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,7 +36,7 @@
 #define SLAPD_TOOLS
 #include "slap.h"
 #include "lutil.h"
-#include "config.h"
+#include "slap-config.h"
 
 typedef struct gluenode {
 	BackendDB *gn_be;
@@ -93,6 +93,12 @@ typedef struct glue_state {
 	int nctrls;
 } glue_state;
 
+typedef struct glue_entry {
+	BackendDB *ge_be;
+	BackendInfo *ge_bi;
+	void *ge_orig;
+} glue_entry;
+
 static int
 glue_op_cleanup( Operation *op, SlapReply *rs )
 {
@@ -225,8 +231,7 @@ glue_op_func ( Operation *op, SlapReply *rs )
 {
 	slap_overinst	*on = (slap_overinst *)op->o_bd->bd_info;
 	BackendDB *b0 = op->o_bd;
-	BackendInfo *bi0 = op->o_bd->bd_info;
-	BI_op_modify **func;
+	BackendInfo *bi0 = op->o_bd->bd_info, *bi1;
 	slap_operation_t which = op_bind;
 	int rc;
 
@@ -247,11 +252,9 @@ glue_op_func ( Operation *op, SlapReply *rs )
 	default: assert( 0 ); break;
 	}
 
-	func = &op->o_bd->bd_info->bi_op_bind;
-	if ( func[which] )
-		rc = func[which]( op, rs );
-	else
-		rc = SLAP_CB_BYPASS;
+	bi1 = op->o_bd->bd_info;
+	rc = (&bi1->bi_op_bind)[ which ] ?
+		(&bi1->bi_op_bind)[ which ]( op, rs ) : SLAP_CB_BYPASS;
 
 	op->o_bd = b0;
 	op->o_bd->bd_info = bi0;
@@ -366,6 +369,8 @@ glue_sub_search( Operation *op, SlapReply *rs, BackendDB *b0,
 		BackendInfo *bi = op->o_bd->bd_info;
 		int rc = SLAP_CB_CONTINUE;
 		for ( on=on->on_next; on; on=on->on_next ) {
+			if ( on->on_bi.bi_flags & SLAPO_BFLAG_DISABLED )
+				continue;
 			op->o_bd->bd_info = (BackendInfo *)on;
 			if ( on->on_bi.bi_op_search ) {
 				rc = on->on_bi.bi_op_search( op, rs );
@@ -547,6 +552,11 @@ glue_op_search ( Operation *op, SlapReply *rs )
 				rs->sr_err = glue_sub_search( op, rs, b0, on );
 			}
 
+			if ( rs->sr_err == SLAPD_NO_REPLY ) {
+				gs.err = rs->sr_err;
+				break;
+			}
+
 			switch ( gs.err ) {
 
 			/*
@@ -713,7 +723,8 @@ end_of_loop:;
 	}
 	rs->sr_ctrls = gs.ctrls;
 
-	send_ldap_result( op, rs );
+	if ( rs->sr_err != SLAPD_NO_REPLY )
+		send_ldap_result( op, rs );
 
 	op->o_bd = b0;
 	op->o_bd->bd_info = bi0;
@@ -785,7 +796,7 @@ glue_tool_entry_open (
 				if ( id != NOID ) {
 					Debug( LDAP_DEBUG_ANY,
 						"glue_tool_entry_open: subordinate database suffix entry DN=\"%s\" also present in superior database rooted at DN=\"%s\"\n",
-						gi->gi_n[i].gn_be->be_suffix[0].bv_val, bd->be_suffix[0].bv_val, 0 );
+						gi->gi_n[i].gn_be->be_suffix[0].bv_val, bd->be_suffix[0].bv_val );
 					return LDAP_OTHER;
 				}
 			}
@@ -910,10 +921,35 @@ glue_entry_get_rw (
 
 	if ( op->o_bd->be_fetch ) {
 		rc = op->o_bd->be_fetch( op, dn, oc, ad, rw, e );
+		if ( rc == LDAP_SUCCESS && *e ) {
+			glue_entry *ge = op->o_tmpcalloc( 1, sizeof(glue_entry),
+					op->o_tmpmemctx );
+
+			if ( !ge ) {
+				if ( op->o_bd->be_release ) {
+					op->o_bd->be_release( op, *e, rw );
+				} else {
+					entry_free( *e );
+				}
+				rc = LDAP_OTHER;
+				goto out;
+			}
+
+			/* b0 is on overlay_entry_get_ov's stack, we'll be passed a fresh
+			 * one at release time */
+			if ( op->o_bd != b0 ) {
+				ge->ge_be = op->o_bd;
+			}
+			ge->ge_bi = op->o_bd->bd_info;
+			ge->ge_orig = (*e)->e_private;
+			(*e)->e_private = ge;
+		}
 	} else {
 		rc = LDAP_UNWILLING_TO_PERFORM;
 	}
-	op->o_bd =b0;
+
+out:
+	op->o_bd = b0;
 	return rc;
 }
 
@@ -924,10 +960,22 @@ glue_entry_release_rw (
 	int rw
 )
 {
+	glue_entry *ge = e->e_private;
 	BackendDB *b0 = op->o_bd;
 	int rc = -1;
 
-	op->o_bd = glue_back_select (b0, &e->e_nname);
+	if ( glueBack ) {
+		op->o_bd = glueBack;
+	} else if ( ge ) {
+		assert( ge->ge_bi != NULL );
+		if ( ge->ge_be )
+			op->o_bd = ge->ge_be;
+		op->o_bd->bd_info = ge->ge_bi;
+		e->e_private = ge->ge_orig;
+		op->o_tmpfree( ge, op->o_tmpmemctx );
+	} else {
+		op->o_bd = glue_back_select( b0, &e->e_nname );
+	}
 
 	if ( op->o_bd->be_release ) {
 		rc = op->o_bd->be_release( op, e, rw );
@@ -1271,7 +1319,7 @@ glue_db_init(
 	if ( SLAP_GLUE_SUBORDINATE( be )) {
 		Debug( LDAP_DEBUG_ANY, "glue: backend %s is already subordinate, "
 			"cannot have glue overlay!\n",
-			be->be_suffix[0].bv_val, 0, 0 );
+			be->be_suffix[0].bv_val );
 		return LDAP_OTHER;
 	}
 
@@ -1330,7 +1378,7 @@ glue_db_destroy (
 	glueinfo		*gi = (glueinfo *)on->on_bi.bi_private;
 
 	free (gi);
-	return SLAP_CB_CONTINUE;
+	return 0;
 }
 
 static int
@@ -1384,6 +1432,11 @@ glue_sub_del( BackendDB *b0 )
 				gi->gi_nodes--;
 			}
 		}
+		/* Mark as no longer linked/sub */
+		b0->be_flags &= ~(SLAP_DBFLAG_GLUE_SUBORDINATE|SLAP_DBFLAG_GLUE_LINKED|
+			SLAP_DBFLAG_GLUE_ADVERTISE);
+		b0->be_pcsn_p = &b0->be_pcsn_st;
+		break;
 	}
 	if ( be == NULL )
 		rc = LDAP_NO_SUCH_OBJECT;
@@ -1443,12 +1496,13 @@ glue_sub_attach( int online )
 				&gi->gi_n[gi->gi_nodes].gn_pdn );
 			gi->gi_nodes++;
 			on->on_bi.bi_private = gi;
+			ga->ga_be->be_pcsn_p = be->be_pcsn_p;
 			ga->ga_be->be_flags |= SLAP_DBFLAG_GLUE_LINKED;
 			break;
 		}
 		if ( !be ) {
 			Debug( LDAP_DEBUG_ANY, "glue: no superior found for sub %s!\n",
-				ga->ga_be->be_suffix[0].bv_val, 0, 0 );
+				ga->ga_be->be_suffix[0].bv_val );
 			/* allow this for now, assume a superior will
 			 * be added later
 			 */
@@ -1479,7 +1533,7 @@ glue_sub_add( BackendDB *be, int advert, int online )
 	if ( overlay_is_inst( be, "glue" )) {
 		Debug( LDAP_DEBUG_ANY, "glue: backend %s already has glue overlay, "
 			"cannot be a subordinate!\n",
-			be->be_suffix[0].bv_val, 0, 0 );
+			be->be_suffix[0].bv_val );
 		return LDAP_OTHER;
 	}
 	SLAP_DBFLAGS( be ) |= SLAP_DBFLAG_GLUE_SUBORDINATE;
